#
#    common.lsf.tcl
#
#    procedures used to LSF (Load Sharing Facility)


#------------------------------------------
# 
# get-jobname
#
# return the name of batch jobs
#
proc get-jobname {prog} {
     return "[string range $prog 0 7]"
}

#------------------------------------------
# 
# berror-mask
#
# Mask the error "lsbatch daemon not responding ... still trying"
#
proc berror-mask {e function} {
   global ERROR _frame

   if {![string first ":" $e]} then {set e [join $e " "]}
   if {[string first "lsbatch daemon not responding" $e] == -1} then {
      clear-busy-watch-error
      midas-warning "$function: $e"
      set-footer "$function: $e"
      return $ERROR
   } else {
     return $e
   }
}

#------------------------------------------
# 
# Ping the Load Information Manager deamon
#
# return  1 if system is OK otherwise 0 
#
proc bping {} {
    global env

    set file    $env(MIDASTempDir)/bping.lsf
    set bpeek  $env(LSFHOME)/bin/bpeek
    set z [catch {exec bpeek -V >& $file &} pid]
    
    set status 0
    if {!$z} then {
       for {set i 0} {$i < 3} {incr i} {
	   if [catch {exec /bin/kill -9 $pid} err] {
	      set info [join [read-file $file] "\n"]
	      if {[lsearch $info "Copyright"] >= 0} then {
                 set status 1
              } else {
                 midas-warning "bping: $info"
                 set status 0
                 break
              }
           }
       }
    }
       
    remove-file $file
    return $status
}

#------------------------------------------
# 
# Tracking batch jobs
#
# Format :
# Name  Status From_Host Exec_Host Job_Id Job_Status User Starting_Time
#
# return  batch jobs list or empty list or ERROR if error
#
proc bjobs {queue} {
    global env ERROR OK
    set bjobs_list {}

    set command $env(LSFHOME)/bin/bjobs
    if {[catch {exec $command -u all -q $queue} m]} then {
	set-footer "Batch Jobs: $m"
	return $ERROR
    }
    set ll [split $m "\n"]
    if {[first [first $ll]] != "JOBID"} then {
       set-footer $m; return $ERROR
    }
       
    set p0 "-----"; # status of program (extern call)

    set data [lreplace $ll 0 0]
    foreach d $data {
 	set p1 [first   $d]; # Job Id
	set p2 [second  $d]; # User
	set p3 [third   $d]; # Status
	set p4 [fourth  $d]; # Queue
	set p5 [fifth   $d]; # From Host; # Exec Host
	if {[llength $d] == 9} then {set p6 "-------"; set begin 5} \
			       else {set p6 [sixth $d]; set begin 6}
	set p7 [lindex $d $begin]; # Job Name
	set p8 [lrange $d [expr $begin+1] end]; # Submit Time

#        set file "$env(MIDASTempDir)/bpeek.$p1"
#        if ![file exist $file] {write-file $file ""}
#        set res1 0; set res2 0
#        set res1 [join [read-file $file] ""]
#        write-file $file [bpeek $p1 50]
#        set res2 [join [read-file $file] ""]       
#        if {[string compare $res1 $res2] != 0} then {set p0 "RUN"} \
#        				       else {set p0 "STOP"}
	set line [format \
	    " %-9s %-9s\t %-8s\t %-8s\t %-15s\t %-11s\t %-7s\t %s" \
	    $p7 $p0 $p5 $p6 $p8 $p2 $p3 $p1]

	lappend bjobs_list $line
    }

    return [lsort $bjobs_list]
}

proc bjobs_names {queue} {
    global env ERROR OK
    set names_list {}

    set command $env(LSFHOME)/bin/bjobs
    set z [catch {exec $command -u all -r -q $queue} m]
    if {$z} then {return {}}

    set ll [split $m "\n"]
    if {[first [first $ll]] != "JOBID"} then {return {}}

    set data [lreplace $ll 0 0]
    set name_prev ""; set name ""
    foreach d $data {
	if {[llength $d] == 9} then {set begin 5} else {set begin 6}
 	set name [lindex $d $begin]
        if {$name != $name_prev} then {lappend names_list $name}	   
        set name_prev $name
    }

    return $names_list
}

proc bjobs_id {queue} {
    global env ERROR OK

    set command $env(LSFHOME)/bin/bjobs
    if [catch {exec $command -u all -r -q $queue} out] {return {}}

    set ll [split $out "\n"]
    if {[first [first $ll]] != "JOBID"} then {return {}}

    set id_list {}
    set data [lreplace $ll 0 0]
    foreach d $data {lappend id_list [first $d]}

    return $id_list
}
#------------------------------------------
# 
# Tracking batch jobs
#
# return OK if there is one or more jobs run else return ERROR
#
proc bjobs_run {queue} {
    global env ERROR OK

    set bjobs $env(LSFHOME)/bin/bjobs
    set z [catch {exec $bjobs -u all -r -q $queue} m]
    if {$z} then {
        if {$m == "No running job found in queue <$queue>"} then {return $ERROR}
	set m [berror-mask $m "Batch Jobs Run"]
	if {$m == $ERROR} then {return $ERROR}
    }
    set ll [split $m "\n"]
    if {[first [first $ll]] != "JOBID"} then {
       set-footer $m; return $ERROR
    }
       
    return $OK
}

#------------------------------------------
# 
# Tracking batch jobs
#
# return info about job Id
#
proc bjobs_info {job_id} {
    global env ERROR OK

    set bjobs $env(LSFHOME)/bin/bjobs
    set z [catch {exec $bjobs -l $job_id} m1]
    if {$z} then {
	set m1 [berror-mask $m1 "Batch Jobs Info"]
	if {$m1 == $ERROR} then {return $ERROR}
    }

    set bhist $env(LSFHOME)/bin/bhist 
    set z [catch {exec $bhist $job_id} m2]
    if {$z} then {
	set m2 [berror-mask $m2 "Batch Jobs Info"]
	if {$m2 == $ERROR} then {return $ERROR}
    }

    set info1 [split $m1 "\n"]
    set info2 [split $m2 "\n"]
    if {[first [second $info2]] != "JOBID"} then {
	set info2 ""
    } else {
	set d [third $info2] 
	set p1 [first   $d]
	set p2 [second  $d]
	set p3 [third   $d]
	set p4 [fourth  $d]
	set p5 [fifth   $d]
	set p6 [sixth   $d]
	set p7 [seventh $d]
	set p8 [eighth  $d]
	set p9 [nineth  $d]
	set p10 [tenth  $d]
	set tmp [format "%-6s %s\t %s\t %s\t %-10s %s\t %s\t %s\t %-5s %s" \
	   		 $p1 $p2 $p3 $p4 $p5 $p6 $p7 $p8 $p9 $p10]
	set info2 [lreplace $info2 2 2 $tmp]
    }

    set info {}
    foreach line $info1 {if {$line != {}} {lappend info $line}}
    lappend info {}
    foreach line $info2 {if {$line != {}} {lappend info $line}}

    return $info
}

#------------------------------------------
# 
# Server Host Information
#
# return the status of the server hosts in the lsbatch system
#
proc bhosts {} {
    global env ERROR OK
    set bhosts_list ""

    set command $env(LSFHOME)/bin/bhosts

    set z [catch {exec $command} m]
    if {$z} then {
	set m [berror-mask $m "Batch Hosts"]
	if {$m == $ERROR} then {return $ERROR}
    }

    set ll [split $m "\n"]
    if {[first [first $ll]] != "HOST_NAME"} then {
       set-footer $m; return $ERROR
    }

    set data [lreplace $ll 0 0]
    foreach d $data {
	append bhosts_list "[first $d] ([second $d])    "
    }

    return $bhosts_list
}


#------------------------------------------
# 
# Server Host Information
#
# return the list of server hosts which status is ok
#
proc bhosts_ok {hosts} {
    global env ERROR OK

    set bhosts $env(LSFHOME)/bin/bhosts
    set z [catch {exec $bhosts} m]
    if {$z} then {
	set m [berror-mask $m "Batch Hosts OK"]
	if {$m == $ERROR} then {return $ERROR}
    }

    set ll [split $m "\n"]
    if {[first [first $ll]] != "HOST_NAME"} then {
       set-footer $m; return $ERROR
    }

    set bhosts_ok {}
    set data [lreplace $ll 0 0]
    foreach d $data {
	if {[second $d] == "ok" } then {lappend bhosts_ok [first $d]}
    }

    foreach host $hosts {
       if {[lsearch -exact $bhosts_ok $host] == -1} then {
	  set-footer "The host \"$host\" is not OK in the lsbatch \
	     	      system !!! "
	  return $ERROR 
       }
    }

    return $OK
}

#------------------------------------------
# 
# Submit a job to the lsbatch system
#
# return  message job or ERROR if error
#
proc bsub {queue machine job output lfile rfile cmd} {
   global env ERROR OK 

   set bsub  $env(LSFHOME)/bin/bsub

   if {$output != ""} then {
      set out "$env(MIDASTempDir)/$output.out"
      set err "$env(MIDASTempDir)/$output.err"
   } else {
      set out "/dev/null"
      set err "/dev/null"
   }
   
   if {$lfile == ""} then {
       set z [catch {exec $bsub -q $queue -m $machine -J $job -o $out -e $err $cmd} m]
   }  else {
       set z [catch {exec $bsub -q $queue -m $machine -J $job -o $out -e $err -f "$lfile > $rfile" $cmd} m]
   }

   if {$z} then {
       set m [berror-mask $m "Batch Submit"]
       if {$m == $ERROR} then {return $ERROR}
   }

   return [split $m "\n"]
}

#------------------------------------------
# 
# Submit a job to the lsbatch system
#
# return  message job or ERROR if error
#
proc bsub-shell {queue job output lfile rfile cmd} {
   global env ERROR OK 

   set bsub  $env(LSFHOME)/bin/bsub

   if {$output != ""} then {
      set out "$env(MIDASTempDir)/$output.out"
      set err "$env(MIDASTempDir)/$output.err"
   } else {
      set out "/dev/null"
      set err "/dev/null"
   }
   			       
   if {$lfile == ""} then {
       set z [catch {exec $bsub -q $queue -J $job -o $out -e $err < ${cmd}} m]
   } else {
       set z [catch {exec $bsub -q $queue -J $job -o $out -e $err -f "$lfile > $rfile" < ${cmd}} m]
   }
         
   if {$z} then {
       set m [berror-mask $m "Batch Submit"]
       if {$m == $ERROR} then {return $ERROR}
   }

   return [split $m "\n"]
} 


#------------------------------------------
# 
# Cancels pending batch and sends signal to running jobs.
# 	SIGQUIT: 3 : start the job (interruption whith core dump, CTRL \)
#	SIGKILL: 2 : kill the job
#	SIGKILL: 9 : (CTRL C)
#	SIGTERM: 15: end of job (kill)
#
# return  message job or ERROR if error
#
proc bkill {queue prog signal} {
   global env ERROR OK
   set messages {}
   set signal 

   if {$queue == "" || $prog == "" || $signal == ""} then {
      set m "Error to send signal to job.\n\
	     bad parameters: \n\
	     queue_name = \"$queue\"\n\
	     signal     = \"$signal\"\n\
	     prog       = \"$prog\""
      midas-error "Bad parameter" $m
      return $ERROR
   }

   set job_name [get-jobname $prog]
   set command $env(LSFHOME)/bin/bkill
   set z [catch {exec $command -s $signal -u all -q $queue -J $job_name 0} out]
   if {$z} then {
	set out [berror-mask $out "Batch Kill"]
	if {$out == $ERROR} then {return $ERROR}
   }

   set title "\[[ttime]\]: Sending signal $signal to neo program \"$prog\"\n"
   set out [linsert [split $out "\n"] 0 $title]

   return [list [join $out "; "]]
}


#------------------------------------------
# 
# Display the stdout and stderr output of a unfinished batch job in 
# the lsbatch system
#
# return ERROR if error
#
proc bpeek {job_id elements} {
    global env ERROR OK

    set command $env(LSFHOME)/bin/bpeek
    if {[catch {exec $command $job_id} out]} then {
	set out [berror-mask $out "Batch Peek"]
	if {$out == $ERROR} then {return $ERROR}
    } else {
#        set last [llength $out]
#        set first [expr $last - $elements]
#        set out [lrange $out $first $last]
#	return [split $out "\n"]
         return $out
    }
}

#------------------------------------------
# 
# Copies a single file from one host to another host using LSF
#
# return  message job or ERROR if error
#
proc lsrcp {source_file dest_host dest_file } {
   global env ERROR OK 

   set lsrcp  $env(LSFHOME)/bin/lsrcp
   set z [catch {exec $lsrcp $source_file @$dest_host:$dest_file} m]
   if {$z} then {
       set m [berror-mask $m "LSF Remote Copy"]
       if {$m == $ERROR} then {return $ERROR}
   }

   return [split $m "\n"]
}

#------------------------------------------
# 
# Launches  make  for a  neo++ program
#
# return pid (of the cmdtool) if no errors otherwise ERROR
#
proc lsmake {host} {
   global env ERROR OK
   
   if {$host == ""} then {
      set-footer "lsmake: host invalid: \"$host\" !!!"
      return $ERROR
   }
   
   set lsmake  $env(LSFHOME)/bin/lsmake
   set cmdtool $env(OPENWINHOME)/bin/cmdtool

   if {[bhosts_ok $host] == $ERROR} then {return $ERROR}

   if {[catch {exec $cmdtool -I "$lsmake -m $host clean; exit" &} pid]} then {
      midas-information $pid
      set-footer "lsmake: make clean failed !!!"
      return $ERROR
   } 
   set i 0
   while {[pidexist $pid]} {incr i; if {$i > 300000} break}
   set pid 0
   
   if {[catch {exec $cmdtool -I "$lsmake -m $host" &} pid]} then {
      midas-information $pid
      set-footer "lsmake: make failed !!!"
      return $ERROR
   }
     
   return $pid
}

#------------------------------------------
# 
# Launches  make  for a  neo++ program
#
# return pid (of the cmdtool) if no errors otherwise ERROR
#     This version uses RSH since no lsf service now in eb09
#
proc lsmake {host} {
   global env ERROR OK
   
   if {$host == ""} then {
      set-footer "lsmake: host invalid: \"$host\" !!!"
      return $ERROR
   }
   
#   set lsmake  $env(LSFHOME)/bin/lsmake
   set cmdtool $env(OPENWINHOME)/bin/cmdtool

   if {[catch {exec $cmdtool -I "rsh $host \"cd [pwd]; make clean\"; exit" &} pid]} then {
      midas-information $pid
      set-footer "lsmake: make clean failed !!!"
      return $ERROR
   } 
   set i 0
   while {[pidexist $pid]} {incr i; if {$i > 300000} break}
   set pid 0
   
   if {[catch {exec $cmdtool -I "rsh $host \"cd [pwd]; make\"" &} pid]} then {
      midas-information $pid
      set-footer "lsmake: make failed !!!"
      return $ERROR
   }
     
   return $pid
}

   

#------------------------------------------
# 
# Graphical user interface for cluster load monitoring
#
#
#
proc xlsmon {} {
   global  env ERROR OK
   
   set-footer "Dispaying GUI for LSF monitoring system..."
   set xlsmon  $env(LSFHOME)/bin/xlsmon

   if {[catch {exec $xlsmon &} pid]} then {
      midas-information $pid
      set-footer "xlsmon: display failed !!!"
      return $ERROR
   }
      
   return $pid
}

